/*
 * Copyright (C) 2004-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include <sys/time.h> /* gettimeofday */
#include "para.h"
#include "gcc-compat.h"
#include <regex.h>
#include <pwd.h>
#include "error.h"

/**
 * para_realloc - paraslash's version of realloc()
 * @p: pointer to the memory block, may be NULL
 * @size: desired new size
 * Description: A wrapper for realloc(). It calls exit(EXIT_FAILURE) on errors,
 * i.e. there is no need to check the return value in the caller.
 */
__must_check __malloc void *para_realloc(void *p, size_t size)
{
	/*
	 * No need to check for NULL pointers: If p is NULL, the  call
	 * to realloc is equivalent to malloc(size)
	 */
	if (!(p = realloc(p, size))) {
		PARA_EMERG_LOG("%s", "realloc failed, aborting\n");
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * para_malloc - paraslash's version of malloc()
 * @size: desired new size
 * Description: A wrapper for malloc(). It calls exit(EXIT_FAILURE) on
 * errors, i.e. there is no need to check the return value in the caller.
 */
__must_check __malloc void *para_malloc(size_t size)
{
	void *p = malloc(size);

	if (!p) {
		PARA_EMERG_LOG("%s", "malloc failed, aborting\n");
		exit(EXIT_FAILURE);
	}
	return p;
}

/**
 * para_calloc - paraslash's version of calloc()
 * @size: desired new size
 * Description: A wrapper for calloc(). It calls exit(EXIT_FAILURE) on
 * errors, i.e. there is no need to check the return value in the caller.
 */
__must_check __malloc void *para_calloc(size_t size)
{
	void *ret = para_malloc(size);

	memset(ret, 0, size);
	return ret;
}

/**
 * para_strdup - paraslash's version of strdup()
 * @s: string to be duplicated
 * Description: A wrapper for strdup(). It calls exit(EXIT_FAILURE) on
 * errors, i.e. there is no need to check the return value in the caller.
 * Moreover, this wrapper checks for @s being NULL and returns an empty
 * string in this case.
 */
__must_check __malloc char *para_strdup(const char *s)
{
	char *ret;

	if ((ret = strdup(s? s: "")))
		return ret;
	PARA_EMERG_LOG("%s", "strdup failed, aborting\n");
	exit(EXIT_FAILURE);
}

/**
 * make_message - allocate a sufficiently large string and print into it
 * @fmt: usual format string
 * Description: Produce output according to @format. No artificial bound
 * on the length of the resuting string is imposed, see printf(3). This
 * function either returns a pointer to a string that must be freed by
 * the caller or aborts without returning.
 */
__must_check __printf_1_2 __malloc char *make_message(const char *fmt, ...)
{
	char *msg;

	PARA_VSPRINTF(fmt, msg);
	return msg;
}

/**
 * para_strcat - paraslash's version of strcat()
 * @a: string to be appended to
 * @b: string to append
 * Description: Append @b to @a. If @a is NULL, return a copy of @b, i.e.
 * para_strcat(NULL, b) is equivalent to para_strdup(b).  If @b is NULL, return
 * @a without making a copy of @a.  Otherwise, construct the concatenation @c,
 * free @a (but not @b) and return @c.
 */
__must_check __malloc char *para_strcat(char *a, const char *b)
{
	char *tmp;

	if (!a)
		return para_strdup(b);
	if (!b)
		return a;
	tmp = make_message("%s%s", a, b);
	free(a);
	return tmp;
}

/**
 * para_dirname - paraslash's version of dirname()
 * @name: Pointer to The full path
 * Description: If @name is NULL or the empty string, return NULL, Otherwise,
 * Make a copy of @name and return its directory component. Caller is
 * responsible to free the result.
 */
__must_check __malloc char *para_dirname(const char *name)
{
	char *p, *ret;

	if (!name || !*name)
		return NULL;
	ret = para_strdup(name);
	p = strrchr(ret, '/');
	if (!p)
		*ret = '\0';
	else
		*p = '\0';
	return ret;
}

/**
 * para_basename - paraslash's version of basename()
 * @name: Pointer to The full path
 * Description: If @name is NULL or the empty string, return NULL, Otherwise,
 * Make a copy of @name and return its filename component. Caller is
 * responsible to free the result.
 */
__must_check __malloc char *para_basename(const char *name)
{
	char *p;

	if (!name || !*name)
		return NULL;
	p = strrchr(name, '/');
	if (!p)
		return para_strdup(name);
	p++;
	if (!*p)
		return NULL;
	return para_strdup(p);
}

/**
 * s_a_r - simple search and replace routine
 * @src: source string
 * @regex: a regular expression
 * @replacement: the replacement text
 * Description: Return a string in which all matches of @regex in @src
 * are replaced by @replacement. Caller must free the result.
 */
__must_check __malloc char *s_a_r(const char *src, const char* regex, const char *replacement)
{
	regex_t preg;
	size_t  nmatch = 1;
	regmatch_t pmatch[1];
	int eflags = 0;
	char *dest = NULL;
	const char *bufptr = src;

//	PARA_DEBUG_LOG("%s: src: %s  regex: %s repl: %s\n", __func__, src,
//		regex, replacement);
	if (!regex || !replacement || !src)
		return para_strdup(src);
	regcomp(&preg, regex, 0);
	while (regexec(&preg,  bufptr, nmatch, pmatch, eflags)
			!= REG_NOMATCH) {
		char *tmp, *arg, *o_bracket, *c_bracket;

		o_bracket = strchr(bufptr + pmatch[0].rm_so, '(');
		c_bracket = o_bracket? strchr(o_bracket, ')') : NULL;
		if (!c_bracket)
			goto out;
		tmp = para_strdup(bufptr);
		tmp[pmatch[0].rm_so] = '\0';
		dest = para_strcat(dest, tmp);
		free(tmp);

		arg = para_strdup(o_bracket + 1);
		arg[c_bracket - o_bracket - 1] = '\0';
		tmp = make_message(replacement, arg);
		free(arg);
		dest = para_strcat(dest, tmp);
		free(tmp);
		bufptr = c_bracket;
		bufptr++;
	}
	dest = para_strcat(dest, bufptr);
//	PARA_DEBUG_LOG("%s: returning %s\n", __func__, dest);
out:
	regfree(&preg);
	return dest;
}

/**
 * s_a_r_list - search and replace multiple regular expressions
 * @regex_list: The array containing a regex/replacement pairs.
 * @src: the source string
 * Description: This function just calls s_a_r() for each element of
 * @regex_list.
 */
__must_check __malloc char *s_a_r_list(struct regex_list *regex_list, char *src)
{
	struct regex_list *rp = regex_list;
	char *ret = NULL, *tmp = para_strdup(src);

	while (rp->regex) {
		ret = s_a_r(tmp, rp->regex, rp->replacement);
		free(tmp);
		if (!ret)
			return src;
		tmp = ret;
		rp++;
	}
	//PARA_DEBUG_LOG("%s: returning %s\n", __func__, dest);
	return ret;
}

/**
 * chop - cut trailing newline
 * @buf: The string to be chopped.
 * Description: Replace the last character in @buf by zero if it is euqal to
 * the newline character.
 */
void chop(char *buf)
{
	int n = strlen(buf);
	if (!n)
		return;
	if (buf[n - 1] == '\n')
		buf[n - 1] = '\0';
}

/**
 * para_tmpname - get a random filename
 * Description: This is by no means a secure way to create temporary
 * files in a hostile direcory like /tmp. However, it is OK to use for
 * temp files, fifos, sockets that are created in ~/.paraslash. Result
 * must be freed by the caller.
 */
__must_check __malloc char *para_tmpname(void)
{
	struct timeval now;
	gettimeofday(&now, NULL);
	srand(now.tv_usec);
	return make_message("%08i", rand());
}

/**
 * para_mkstemp - create unique temporary file
 * @template: the template to be passed to mkstemp()
 * @mode: The desired mode of the tempfile
 * Description: This wrapper for mkstemp additionally uses fchmod() to
 * set the given mode of the tempfile if mkstemp() returned success.
 * Return value: The file descriptor of the temp file just created on success.
 * On errors, -E_MKSTEMP or -E_FCHMOD is returned.
 */
__must_check int para_mkstemp(char *template, mode_t mode)
{
	int tmp, fd = mkstemp(template);

	if (fd < 0)
		return -E_MKSTEMP;
	tmp = fchmod(fd, mode);
	if (tmp >= 0)
		return fd;
	close(fd);
	unlink(template);
	return -E_FCHMOD;
}

/**
 * para_logname - get the logname of the current user
 *
 * Returns a dynammically allocated string that must be freed by
 * the caller. On errors, the string "unknown user" is returned,
 * i.e. this function never returns NULL.
 */
__must_check __malloc char *para_logname(void)
{
	struct passwd *pw = getpwuid(getuid());
	return para_strdup(pw? pw->pw_name : "unknown_user");
}

/**
 * para_homedir - get the home directory of the current user
 *
 * Returns a dynammically allocated string that must be freed by the
 * caller. If the home directory could not be found, this function
 * returns "/tmp".
 */
__must_check __malloc char *para_homedir(void)
{
	struct passwd *pw = getpwuid(getuid());
	return para_strdup(pw? pw->pw_dir : "/tmp");
}

/**
 * split_args -- split string and return pointers to its parts.
 * @args: The string to be split
 * @argv_ptr: pointer to the list of substrings
 * @delim: delimiter
 *
 * This function modifies @args by replacing each occurance of @delim
 * by zero. A NULL-terminated array of pointers to char* is allocated
 * dynamically and these pointers are initialized to point to the broken-up
 * substrings within @args. A pointer to this array is returned via @argv_ptr.
 * Return value: The number of substrings found in @args.
 */
__must_check unsigned split_args(char *args, char ***argv_ptr, int delim)
{
	char *p = args;
	char **argv;
	ssize_t n = 0, i;

	while ((p = strchr(p, delim))) {
		p++;
		n++;
	}
	*argv_ptr = para_malloc((n + 3) * sizeof(char *));
	argv = *argv_ptr;
	i = 0;
	p = args;
//	printf("split_args: a:%s\n", p);
	while (p) {
		argv[i] = p;
		p = strchr(p, delim);
		if (p) {
//		printf("a:%s\n", p);
			*p = '\0';
//		printf("b:\n");
			p++;
			i++;
		}
	}
	argv[n + 1] = NULL;
	return n;
}

/**
 * valid_fd_012 - ensure that file descriptors 0, 1, and 2 are valid
 *
 * Common approach that opens /dev/null until it gets a file descriptor greater
 * than two, see okir's Black Hats Manual.
 */
void valid_fd_012(void)
{
	while (1) {
	int     fd;

		fd = open("/dev/null", O_RDWR);
		if (fd < 0)
			exit(EXIT_FAILURE);
		if (fd > 2) {
			close(fd);
			break;
		}
	}
}
